#pragma once
#include <memory>
#include <vector>
#include <brick-types>

using brick::types::Union;

template< typename T >
using Ptr = std::shared_ptr< T >;

struct Expression;

/* ... */

struct IfLike
{
    Ptr< Expression > condition, body;
};

struct If : IfLike {};
struct While : IfLike {};

/* ... */

using ExpBase = Union< Let, If, While, App, Call, Identifier, Constant >;
struct Expression : ExpBase { explicit Expression( ExpBase b ) : ExpBase( b ) {} };
using Global = Union< Def, Let >;

struct Toplevel
{
    std::vector< Global > globals;
};

namespace {

/* ... */

std::ostream &operator<<( std::ostream &o, Expression e )
{
    e.match( [&]( If v ) { o << "if?"; },
             [&]( Call v ) { o << "(" << v.fun << AppLike( v ) << ")"; },
             [&]( App v ) { o << "(" << v.op << AppLike( v ) << ")"; },
             [&]( While v ) { o << "(while" << IfLike( v ) << ")"; },
             [&]( Identifier i ) { o << i; },
             [&]( Constant c )
             {
                 c.match( [&]( int i ) { o << i; },
                          [&]( std::string s ) { o << s; } );
             },
             [&]( Let v ) { o << "let?"; } );
    return o;
}

/* ... */

}
