#pragma once
#include <fstream>
#include <map>

struct Token
{
    enum Cat
    {
        ParenOpen, ParenClose,
        LitString, Identifier,
        /* ... */
        Error, Eof
    } cat;

    std::string text;
    int line;

    explicit operator bool() { return cat != Eof; }
    Token( Cat cat, std::string text, int line )
        : cat( cat ), text( text ), line( line )
    {}
};

static const char *catnames[] = { /* ... */ };

static inline std::ostream &operator<<( std::ostream &o, Token t )
{
    return o << "[" << catnames[ t.cat ] << " " << t.text << " at line " << t.line << "]";
}

struct Lexer
{
    std::map< char, Token::Cat > operators;
    std::map< std::string, Token::Cat > keywords;

    std::ifstream in;
    std::string buf;
    char c;
    int line = 1;

    Lexer( const char *f ) : in( f )
    {
        /* ... set up operators & keywords */
    }

    Token next()
    {
        whitespace();
        buf += ( c = in.get() );

        if ( c == EOF )
            return Token( Token::Eof, "", line );

        if ( c == '(' )
            return shift( Token::ParenOpen );

        /* ... */

        return shift( Token::Error );
    }

protected:

    void whitespace()
    {
        while ( std::isspace( c = in.get() ) )
            if ( c == '\n' )
                ++ line;
        in.unget();
    }

    Token shift( Token::Cat c )
    {
        Token t( c, buf, line );
        buf.clear();
        return t;
    }

    /* ... */
};
