#pragma once
#include "lexer.hpp"
#include "ast.hpp"

struct BadParse
{
    Token found;
    std::string expected;
    BadParse( Token t, std::string e ) : found( t ), expected( e ) {}
};

static inline std::ostream &operator<<( std::ostream &o, BadParse bp )
{
    return o << "parse error at " << bp.found << ", expected " << bp.expected;
}

struct Parser
{
    Lexer lexer;
    Token tok;

    Parser( const char *file ) : lexer( file ), tok( Token::Eof, "", 0 ) {}

    template< typename T >
    Ptr< Expression > make_expr( T e ) { return std::make_shared< Expression >( e ); }

    void shift() { tok = lexer.next(); }

    void fail( std::string err ) __attribute__((noreturn))
    {
        throw BadParse( tok, err );
    };

    /* recursive descent */

    Toplevel toplevel()
    {
        Toplevel result;
        while ( true )
        {
            shift();
            if ( tok.cat == Token::Eof )
                return result;
            result.globals.push_back( global() );
        }
    }

    Identifier identifier()
    {
        if ( tok.cat != Token::Identifier )
            fail( "identifier" );
        Identifier id;
        id.name = tok.text;
        return id;
    }

    /* ... */
};
